/***************************************************************************************************
*            Sample for the usage of TopoART-C (classes TopoART_C, Fast_TopoART_C) and             *
*            Hypersphere TopoART-C (class Hypersphere_TopoART_C)                                   *
****************************************************************************************************
*                            Created by Marko Tscherepanow, 17 April 2017                          *
***************************************************************************************************/

// Compile and run from the console: dotnet run TopoART-C_sample2

using System;
using System.IO;
using System.Reflection;
using System.Text.RegularExpressions;
using LibTopoART;
using System.Globalization;

namespace LibTopoART_samples
{
	/// <summary>
	/// Classification sample using more complex synthetic two-dimensional data with associated class IDs.
	/// <para>
	/// Train TopoART-C with a two-dimensional dataset similar to the one used in "Marko Tscherepanow and Sören Riechers
	/// (2012). An Incremental On-line Classifier for Imbalanced, Incomplete, and Noisy Data. In Proceedings of the 
	/// European Conference on Artificial Intelligence (ECAI), Workshop on Active and Incremental Learning (AIL),
	/// pp. 18-23. Montpellier, France." This dataset comprises six clusters (each containing 15,000 samples) as well as
	/// 10,000 noise samples which were mixed randomly. The samples are divided into two classes. Each sample belonging 
	/// to one of the six clusters is assigned a class ID depending on its position. In contrast, noise samples receive 
	/// a random class ID. 
	/// </para>
	/// <para>The resulting neural network can be visualised using the R script <c>ShowTopoARTCResults.R</c> 
	/// or the R script <c>ShowHypersphereTopoARTCResults.R</c>, respectively. Both R scripts are provided 
	/// in the subfolder <c>R</c>.
	/// </para>
	/// </summary>
	class TopoART_C_sample2
	{
		enum Selector {
			USE_TopoART_C,
			USE_Fast_TopoART_C,
			USE_Hypersphere_TopoART_C
		};

	    static void Main() {
	
			// Dataset
			const string Dataset 		=	"../../../../../data/AIL12-like_dataset.txt";		
			// Destination directory for trained networks
			const string NetworkPath	=	"../../../../../results/networks/";

			const long SampleNumber		=	100000;
			const long InputDimension	=	2;

			// Choose TopoART implementation (see enum Selector above)
			Selector sel = Selector.USE_Fast_TopoART_C;

			ITopoART_C tac			=	null;
			string FilePrefix		=	null;
			string FileSuffix		=	null;
			decimal[][] InputArray	=	new decimal[SampleNumber][];
			long[] ClassIDArray		=	new long[SampleNumber];
			StreamReader DatasetFile;
			long i;
			string line;
			DateTime start, end;
			TimeSpan time;
			Random rnd = new Random();

			// Set working directory to assembly directory
            Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().CodeBase).LocalPath));

			// Start time measuring
			start = DateTime.Now;
			
			// Load dataset
			using(DatasetFile = new StreamReader(File.Open(Dataset, FileMode.Open))) {
				for(i = 0; i < SampleNumber; ++i) {
					line = DatasetFile.ReadLine();
					string[] entries = Regex.Split(line, @"\s+");
					InputArray[i] = new decimal[InputDimension]; 
					InputArray[i][0] = Decimal.Parse(entries[1], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);
					InputArray[i][1] = Decimal.Parse(entries[2], NumberStyles.Float, System.Globalization.CultureInfo.InvariantCulture);

					// Compute class IDs of cluster samples depending on their x coordinate
					if(entries[3] != "NOISE")
						ClassIDArray[i] = (InputArray[i][0] > 0.5m) ? 1 : 2;
					// Assign a random class ID to noise samples
					else
						ClassIDArray[i] = rnd.Next(1, 3);		
				}
			}
			
			// Initialise a TopoART network with appropriate parameter values
			switch(sel) {
				case Selector.USE_TopoART_C:
						tac = new TopoART_C(InputDimension, 2, 0.92m);	
						tac.Beta_sbm	=	0.65m;
						tac.Phi			=	5;
						tac.Tau			=	200;
						FilePrefix 		=	"TopoART-C";
						FileSuffix		=	"tac";
					break;
				case Selector.USE_Fast_TopoART_C:
						tac = new Fast_TopoART_C(InputDimension, 2, 0.92m);	
						tac.Beta_sbm	=	0.65m;
						tac.Phi			=	5;
						tac.Tau			=	200;
						FilePrefix 		=	"Fast_TopoART-C";
						FileSuffix		=	"ftac";
					break;
				case Selector.USE_Hypersphere_TopoART_C:
					tac = new Hypersphere_TopoART_C(InputDimension, 2, 0.90m);	
						tac.Beta_sbm	=	0.60m;
						tac.Phi			=	5;
						tac.Tau			=	200;
						FilePrefix 		=	"Hypersphere_TopoART-C";
						FileSuffix		=	"htac";
					break;
			}

			// Train network
			for(i = 0; i < SampleNumber; ++i) {
				tac.Learn(InputArray[i], ClassIDArray[i]);
				if(i % 1000 == 0) 
					Console.Write(".");
			}
			Console.WriteLine("");
			
			// Determine clusters
			tac.ComputeClusterIDs();
			
			// Save network in human readable form
			tac.SaveText(NetworkPath + FilePrefix + "_AIL12-like_dataset.txt");

			// Save network in binary form
			tac.Save(NetworkPath + FilePrefix + "_AIL12-like_dataset." + FileSuffix);

			// Stop time measuring
			end = DateTime.Now;

			// Output the required time
			time = end - start;
			Console.WriteLine("time for computation: " + time);
		}
	}
}